package Pederson;

/*Extracting all non-duplicated paths from XML document
 *Insert the paths into a "path" table in MySQL database
*/

import java.io.*;
import java.util.Stack;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;


public class MakeCovering extends DefaultHandler
{
    String inFileName; 
    String outFileName; 
    String root = "sales";
    String fact = "sale";
    String[] dim = {"part", "customer", "supplier", "date"};
    String[][] subDim = {
                            {"type3", "type2", "type1"},
                            {"nation", "region"},
                            {"nation", "region"},
                            {"day", "month", "year"}
                        };
    
    String[] measure = {"f_quantity", "f_totalamount"};
    Stack elementStack;
    int dimID;
    int subDimID;
    static File wf;
    String outSubDim[];
    String outMeasure[];
    String outFact;
    static BufferedWriter bw;
    String mctxt;
    
    public void genMakeCovering () throws Exception
    {
        mctxt = "MakeCovering: ";
        long rtime = System.currentTimeMillis();
        
        inFileName = new java.io.File("").getAbsolutePath() + "\\outputDWXML\\f_sale.xml"; 
        outFileName = new java.io.File("").getAbsolutePath() + "\\PedersenOverheadXML\\complete_sale.xml";
        
	XMLReader xr = XMLReaderFactory.createXMLReader();
	MakeCovering handler = new MakeCovering();
	xr.setContentHandler(handler);
	xr.setErrorHandler(handler);

        //Input file
        wf = new File(outFileName);
        try {
            bw = new BufferedWriter(new FileWriter (wf, false));
        } catch (IOException ex) {
            Logger.getLogger(MakeCovering.class.getName()).log(Level.SEVERE, null, ex);
        }
        FileReader r = new FileReader(inFileName);
        xr.parse(new InputSource(r));
        
        rtime = System.currentTimeMillis() - rtime;
             
        mctxt = mctxt + "Running time: " + rtime + " milliseconds.";
    }

    ////////////////////////////////////////////////////////////////////
    // Event handlers.
    ////////////////////////////////////////////////////////////////////


    @Override
    public void startDocument ()
    {        
        //initializating
        elementStack = new Stack();
        dimID = -1;
        subDimID = -1; 
        outFact = "";
        try {
            bw = new BufferedWriter(new FileWriter (wf, true));
        } catch (IOException ex) {
            Logger.getLogger(MakeCovering.class.getName()).log(Level.SEVERE, null, ex);
        }
    }


    @Override
    public void endDocument ()
    {
        try {
            bw.close();
        } catch (IOException ex) {
            Logger.getLogger(MakeCovering.class.getName()).log(Level.SEVERE, null, ex);
        }
    }


    @Override
    public void startElement (String uri, String name, String qName, Attributes atts)
    {
        //root
        if(qName.equals(root)){
            try {
                bw.write("<" + qName + ">");
            } catch (IOException ex) {
                Logger.getLogger(MakeCovering.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        //fact
        else if (qName.equals(fact)){
            outFact = "<" + qName + ">";
            outMeasure = new String[measure.length];
        }
        //dimension or measure
        else if (elementStack.peek().equals(fact)){
            dimID = -1;
            //dimension
            int i=0;
            while (i < dim.length){
                if(qName.equals(dim[i])){
                    dimID = i;
                    outFact = outFact 
                            + "\n"
                            + "\t<" + qName + " value = \'" + atts.getValue("value") + "\' >";
                    break;
                }
                i++;
            }
            //measure
            if(dimID < 0){
                int j=0;
                while (j < measure.length){
                    if(qName.equals(measure[j])){
                        outMeasure[j] = atts.getValue("value");
                    }
                    j++;
                }
            }
        }
        //subdimension
        
        else if (dimID >= 0){
            if(elementStack.peek().equals(dim[dimID])){
                outSubDim = new String[subDim[dimID].length];
            }
            int i=0;
            while (i < subDim[dimID].length){
                if (qName.equals(subDim[dimID][i])){
                    outSubDim[i] = atts.getValue("value");
                    break;
                }
                i++;
            }
        }
        
        elementStack.push(qName);
        
    }


    @Override
    public void endElement (String uri, String name, String qName)
    {
        //root
        if(qName.equals(root)){
            try {
                bw.newLine();
                bw.write("</" + qName + ">");
            } catch (IOException ex) {
                Logger.getLogger(MakeCovering.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        
        //fact
        else if(qName.equals(fact)){
            //add measure
            for(int i=0; i<outMeasure.length; i++){
                outFact = outFact 
                        + "\n"
                        + "\t<" + measure[i] + " value = \'" + outMeasure[i] + "\' />";
            }
            
            //add end of fact
            outFact = outFact 
                    + "\n"
                    + "</" + qName + ">";
            try {
                bw.newLine();
                bw.write(outFact);
            } catch (IOException ex) {
                Logger.getLogger(MakeCovering.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        else if(dimID >=0 && qName.equals(dim[dimID])){
             //add end of dimension
            outFact = outFact 
                    + "\n"
                    + "\t</" + qName + ">";
            dimID = -1; 
        }
       
        
        elementStack.pop();
         //subdimension
        if (dimID >= 0 && elementStack.peek().equals(dim[dimID])){
            //add subdimension
            outFact = outFact + "\n\t\t";
            for(int i=0; i < subDim[dimID].length; i++){
                if(outSubDim[i]==null){
                    outFact = outFact 
                        + "<" + subDim[dimID][i] + " value = \'OTHER\' >";
                }
                else{
                outFact = outFact 
                        + "<" + subDim[dimID][i] + " value = \'" + outSubDim[i] + "\' >";
                }
            }
            for(int i= subDim[dimID].length - 1; i>=0; i--){
                outFact = outFact + "</" + subDim[dimID][i] + ">";
            }
        }
    }
}