package QBS;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;


public class QBS extends DefaultHandler
{
    String directory = new java.io.File("").getAbsolutePath();
    String inFileName = directory + "\\outputDWXML\\f_sale.xml";   
    String outFileName = directory + "\\outputGrouping\\QBSGrouping.xml";    
    //Query Pattern
    String QPRoot = "sales";//Query Pattern root
    String QPFact = "sale"; //Query Pattern fact
    String[] QPDimensions; //Query Pattern dimension list
    String[] QPSubDimensions; //Query Pattern subdimension List
    String[] QPMeasures; //suppose that only one measure to be aggregated
    String QPAggFunction; //sum, min, max, or avg
    
    //Withness XML
    String[] outDimensions;//store output of all dimesnions in current fact for checking duplicate subdimension
    String[] outMeasures;//store output of all measures in current fact
    //String currentMeasure;//measure of current fact
    int dimID; //Current dimension id
    int mID; //Current measure id
    Stack elementStack;//Stack for storing the parsed elements
    int hKey;
    static Hashtable hashtable;
    int subDimAdded; // status to check whether the dimension is already added to output
    String QBStxt;    
    
    
    public QBS (String[] dimensions, String[] subDimensions, String[] measures, String aggregation) throws IOException{        
        QPDimensions = dimensions;
        QPSubDimensions = subDimensions;
        QPMeasures = measures;
        QPAggFunction = aggregation;
        String txt = "Starting QBS.\n";
        long rtime = System.currentTimeMillis();   
        
        parseXML(inFileName);
        writeToXML();
        
        rtime = System.currentTimeMillis() - rtime;
        txt = txt + "End QBS.\n";        
        txt = txt + "Running time: " + rtime + " milliseconds.\n";
        txt = txt + "output: " + outFileName;
        QBStxt = txt;
        
    }
    
        
    public void parseXML(String uri)
    {
        DefaultHandler handler = this;
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try {
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse(new File(uri), handler);
        } catch (ParserConfigurationException | SAXException | IOException t) {}
    }
    @Override
    public void startDocument ()
    {        
        //Initializing
        outDimensions= new String[QPDimensions.length];
        outMeasures = new String[QPMeasures.length];
        dimID = -1;
        mID = -1;
        subDimAdded = 0;
        
        elementStack = new Stack();
        hKey = 0;
        hashtable = new Hashtable();
    }
    @Override
    public void startElement (String uri, String name, String qName, Attributes atts)
    {
        //fact
        if(qName.equals(QPFact)){
            outDimensions = new String[QPDimensions.length];
            outMeasures = new String[QPMeasures.length];
        }
        else if (!elementStack.empty()){
            //dimension or measure
            if (elementStack.peek().equals(QPFact)){
                dimID = -1;
                mID = -1;
                
                //Checking for current measure
                for(int i=0; i<QPMeasures.length; i++){
                    if(qName.equals(QPMeasures[i])){
                        mID = i;
                        outMeasures[i] = atts.getValue("value");
                    }
                }
                
                //Checking for current dimension
                int i=0;
                while (i<QPDimensions.length){
                    if(qName.equals(QPDimensions[i])){
                        dimID = i;
                        subDimAdded = 0;
                        break;
                    }
                    i++;
                }
                
            }
            
            //subdimension
            if (dimID >= 0){
                if(qName.equals(QPSubDimensions[dimID]) && subDimAdded == 0){
                    String currentSubDim = atts.getValue("value");
                    addSubDimension(currentSubDim);
                    subDimAdded = 1;
                }
            }
            else if(mID >= 0){
                outMeasures[mID] = atts.getValue("value");
            }
        }
        
        elementStack.push(qName);//add element name to stack
    }
    @Override
    public void endElement (String uri, String name, String qName)
    {   
        if(qName.equals(QPFact)){
            String factText = "";
            for (int i=0; i<outDimensions.length; i++){
                if (i==0){
                    factText = outDimensions[i];
                }
                else{
                    factText = factText + "\t" + outDimensions[i];
                }
            }
            //add current measure
            for(int i=0; i<outMeasures.length; i++){
                factText = factText + "\t" + outMeasures[i];
            }
            writeLine(factText);
        }
        /*else if (dimID >= 0 && qName.equals(QPDimensions[dimID])){
            dimID = -1;
        }*/
        elementStack.pop();
        //if subdimension is missing
        if(!elementStack.empty() && dimID >= 0 && elementStack.peek().equals(QPDimensions[dimID])){
            if (subDimAdded == 0) addSubDimension("OTHER");
            subDimAdded = 0;
            
        }
    } 
    
    public void addSubDimension(String subDimValue){
        if(outDimensions[dimID] == null){
            outDimensions[dimID] = subDimValue;
        }
        else{
            int st = 0;
            StringTokenizer token = new StringTokenizer(outDimensions[dimID], ",");
            while (token.hasMoreTokens()){
                if(subDimValue.equals(token.nextToken())){
                    st = 1;
                }
            }
            if(st == 0){
                outDimensions[dimID] = outDimensions[dimID] + "," + subDimValue;
            }
        }
    }
    public void writeLine(String newGroup){
        //Check duplicated line
        int matchedLineNumber = 0;
        String matchedLineString = "";
        int hk = hKey;
        while (hk > 0){
            String currentLine = (String) hashtable.get(hk);
            matchedLineString = checkMatchedDimensions(newGroup, currentLine);
            if(!matchedLineString.equals("")){
                matchedLineNumber = hk;
                break;
            }
            hk--;
        }
        if(matchedLineNumber>0){
            hashtable.put(matchedLineNumber, matchedLineString);
        }
        else{
            hKey++;
            hashtable.put(hKey, newGroup);
        }
    }
    public String checkMatchedDimensions(String newLine, String currentLine){
        String matchedString = "";
        int st = 0;
        //Get list of dimension from newline and currentLine
        StringTokenizer tokenNewLine = new StringTokenizer(newLine, "\t");
        StringTokenizer tokenCurrentLine = new StringTokenizer(currentLine, "\t");       
        String[] nl = new String[QPDimensions.length + QPMeasures.length];//new line
        String[] cl = new String[QPDimensions.length + QPMeasures.length];//current line
        int i=0;
        while (tokenNewLine.hasMoreTokens()){
            nl[i] = tokenNewLine.nextToken();
            i++;
        }
        i=0;
        while (tokenCurrentLine.hasMoreTokens()){
            cl[i] = tokenCurrentLine.nextToken();
            i++;
        }
        i=0;       
        w1:while(i < QPDimensions.length){
            st=0;
            StringTokenizer tokenNewDimension = new StringTokenizer(nl[i], ",");
            StringTokenizer tokenCurrentDimension = new StringTokenizer(cl[i], ",");       
            
            //if the number of words in both dimensions are equal,  
            //the new dimension and the dimension in output file may be matched
            //Check if all words matched
            int nlWords = tokenNewDimension.countTokens();////number of words in new output dimension
            int clWords = tokenCurrentDimension.countTokens();//number of words in output dimension
            
            if(nlWords==clWords){               
                w2:while (tokenNewDimension.hasMoreTokens()){
                    String ns = tokenNewDimension.nextToken();
                    w3:while(tokenCurrentDimension.hasMoreTokens()){
                        String cs = tokenCurrentDimension.nextToken();
                        if(ns.equals(cs)){
                            st=1;
                            break w3;
                        }
                        else{
                            st=0;
                        }
                    }
                    if (st==0){
                        break w2;
                    }
                }
            }
            if (st == 0){
                break w1;
            }
            else{
                i++;
            }
        }
        if (st ==1){    
            for(int k=0; k<QPDimensions.length; k++){
                if(k==0){
                    matchedString = matchedString + nl[k];
                }
                else{
                    matchedString = matchedString + "\t" + nl[k];
                }
            }
            if(!QPAggFunction.equals("avg")){
                double[] measures = new double[QPMeasures.length];
                if(QPAggFunction.equals("sum")){
                    for (int j=0; j<measures.length; j++){
                        int measureid = QPDimensions.length + j;
                        measures[j] = Double.valueOf(cl[measureid]) + Double.valueOf(nl[measureid]);
                    }
                }
                else if(QPAggFunction.equals("min")){
                    for (int j=0; j<measures.length; j++){
                        int measureid = QPDimensions.length + j;

                        if(Double.valueOf(cl[measureid]) < Double.valueOf(nl[measureid])){
                            measures[j] = Double.valueOf(cl[measureid]);
                        }
                        else {
                            measures[j] = Double.valueOf(nl[measureid]);
                        }
                    }
                }
                else if(QPAggFunction.equals("max")){
                    for (int j=0; j<measures.length; j++){
                        int measureid = QPDimensions.length + j;

                        if(Double.valueOf(cl[measureid]) < Double.valueOf(nl[measureid])){
                            measures[j] = Double.valueOf(nl[measureid]);
                        }
                        else {
                            measures[j] = Double.valueOf(cl[measureid]);
                        }
                    }
                }
                for(int j=0; j<measures.length; j++){
                    matchedString = matchedString + "\t" + measures[j];
                }
            }
            else{//Average
                String[] measures = new String[QPMeasures.length];
                for(int j=0; j<measures.length; j++){
                    int measureid = QPDimensions.length + j;
                    measures[j] = cl[measureid] +  "," + nl[measureid];
                }
                for(int j=0; j<measures.length; j++){
                    matchedString = matchedString + "\t" + measures[j];
                }
            }
        }
        return matchedString;
    }
    public void writeToXML() throws IOException{
        Set set = hashtable.keySet();
        Iterator iterator = set.iterator();
        File wf = new File(outFileName);
        try (BufferedWriter bw = new BufferedWriter(new FileWriter (wf))) {
            bw.write("<?xml version=\'1.0\' encoding=\'UTF-8\'?>\n");
            bw.write("<"+QPRoot+">\n");
            String newLine;  
            while(iterator.hasNext()){
                newLine = (String) hashtable.get(iterator.next());
                bw.write("\t<"+QPFact+">\n");
                StringTokenizer strToken = new StringTokenizer(newLine, "\t");
                int d = 0;
                while(strToken.hasMoreTokens()){
                    //Measures
                    if(d >= QPSubDimensions.length){
                        if(QPAggFunction.equals("avg")){
                            StringTokenizer measureToken = new StringTokenizer(strToken.nextToken(), ",");
                            int m = 0;
                            double mValue = 0;
                            while(measureToken.hasMoreTokens()){
                                mValue = mValue + Double.valueOf(measureToken.nextToken());
                                m++;
                            }
                            mValue = mValue / m;
                            bw.write("\t\t<"+QPMeasures[d-QPSubDimensions.length]+">"+mValue+"</"+QPMeasures[d-QPSubDimensions.length]+">\n");
                        }
                        else{
                            bw.write("\t\t<"+QPMeasures[d-QPSubDimensions.length]+">"+strToken.nextToken()+"</"+QPMeasures[d-QPSubDimensions.length]+">\n");
                        }
                    }
                    //Dimensions
                    else if (d >= 0 && d < QPSubDimensions.length){
                        String dStr = strToken.nextToken();
                        if(dStr == null || dStr.equals("") || dStr.equals("null")){
                            dStr = "OTHER";
                        }
                        bw.write("\t\t<"+QPSubDimensions[d]+">" + dStr + "</"+QPSubDimensions[d]+">\n" );
                    }
                    d++;
                }
                bw.write("\t</"+QPFact+">\n");
            }
            bw.write("</"+QPRoot+">\n");
            bw.close();
        }
    }
}